/*
 *  Project   : tin - a visual threaded usenet newsreader
 *  Module    : art.c
 *  Author    : R.Skrenta / I.Lea
 *  Created   : 01-04-91
 *  Updated   : 29-09-91
 *  Release   : 1.0
 *  Notes     :
 *  Copyright : (c) Copyright 1991 by Rich Skrenta & Iain Lea
 *	            You may  freely  copy or  redistribute  this software,
 *              so  long as there is no profit made from its use, sale
 *              trade or  reproduction.  You may not change this copy-
 *              right notice, and it must be included in any copy made
 */

#include	"tin.h"

#define HEADER_LEN	1024

extern int errno;

char index_file[LEN];
char *glob_art_group;
int index_file_killed = FALSE;
long last_read_article;


/*
 *  Convert a string to a long, only look at first n characters
 */

long my_atol (s, n)
	char *s;
	int n;
{
	long ret = 0;

	while (*s && n--) {
		if (*s >= '0' && *s <= '9')
			ret = ret * 10 + (*s - '0');
		else
			return -1;
		s++;
	}

	return ret;
}


/*
 *  Construct the pointers to the basenotes of each thread
 *  arts[] contains every article in the group.  inthread is
 *  set on each article that is after the first article in the
 *  thread.  Articles which have been expired have their thread
 *  set to -2 (ART_EXPIRED).
 */

void find_base ()
{
	int i;

	top_base = 0;

	for (i = 0; i < top; i++)
		if (! arts[i].inthread && arts[i].thread != ART_EXPIRED) {
			if (top_base >= max_art)
				expand_art ();
			base[top_base++] = i;
		}
}

/* 
 *  Count the number of non-expired and non-killed articles in arts[]
 */

int num_of_arts ()
{
	int sum = 0;
	register int i;

	for (i = 0; i < top; i++) {
		if (arts[i].thread != ART_EXPIRED && ! arts[i].tagged) {
			sum++;
		}
	}

	return sum;
}

/*
 *  Do we have an entry for article art?
 */

int valid_artnum (art)
	long art;
{
	register int i;

	for (i = 0; i < top; i++)
		if (arts[i].artnum == art)
			return i;

	return -1;
}


/*
 *  Return TRUE if arts[] contains any expired articles
 *  (articles we have an entry for which don't have a corresponding
 *   article file in the spool directory)
 */

int purge_needed ()
{
	register int i;

	for (i = 0; i < top; i++)
		if (arts[i].thread == ART_EXPIRED)
			return TRUE;

	return FALSE;
}


/*
 *  Main group indexing routine.  Group should be the name of the
 *  newsgroup, i.e. "comp.unix.amiga".  group_path should be the
 *  same but with the .'s turned into /'s: "comp/unix/amiga"
 *
 *  Will read any existing index, create or incrementally update
 *  the index by looking at the articles in the spool directory,
 *  and attempt to write a new index if necessary.
 */

void index_group (group, group_path)
	char *group;
	char *group_path;
{
	int killed = FALSE;
	int modified = FALSE;
	glob_art_group = group;

#ifdef SIGTSTP
	if (do_sigtstp) {
#ifdef POSIX_JOB_CONTROL
		sigemptyset (&art_act.sa_mask);
		art_act.sa_flags = SA_RESTART | SA_RESETHAND;
		art_act.sa_handler = art_suspend;
		sigaction (SIGTSTP, &art_act, 0L);
#else
		signal (SIGTSTP, art_suspend);
#endif
	}
#endif

#ifdef SIGWINCH
	signal (SIGWINCH, art_resize);
#endif

	if (! update) {
		sprintf (msg, txt_group, group);
		wait_message (msg);
	}

	hash_reclaim ();
	free_art_array ();

	if (local_index)
		find_local_index (group);
	else
		sprintf (index_file, "%s/%s/%s", spooldir, group_path, INDEXDIR);

	/*
	 *  load articles from index file if it exists
	 */
	load_index ();

	/*
	 *  load killed articles into arts[] because kill arts is OFF
	 */
	if (! kill_articles && index_file_killed) {
		index_file_killed = FALSE;
		last_read_article = 0L;
	}

	/*
	 *  add any articles to arts[] that are new or were killed
	 */
	modified = read_group (group, group_path);

	/*
	 *  compare kill descriptions to arts[] and kill mark any that match
	 */
	killed = kill_any_articles (group);
	
	if (modified || killed || purge_needed()) {
		if (local_index) {		/* writing index in home directory */
			set_real_uid_gid ();	/* so become them */
		}

		if (killed) {
			reload_index_file (group, killed);
		} else {
			dump_index (group, FALSE);
			make_threads (FALSE);
			find_base ();
		}

		if (local_index) {
			set_tin_uid_gid ();
		}
	} else {
		make_threads (FALSE);
		find_base ();
	}
	
	if ((modified || killed) && ! update) {
		clear_message();
	}
}

/*
 *  Index a group.  Assumes any existing index has already been
 *  loaded.
 */

int read_group (group, group_path)
	char *group;
	char *group_path;
{
	int fd;
	long art;
	int count;
	int display_groupname = TRUE; 
	int modified = FALSE;
	int respnum;
	register int i;
	
	setup_base (group, group_path);	/* load article numbers into base[] */
	count = 0;

	for (i = 0; i < top_base; i++) {	/* for each article # */
		art = base[i];

/*
 *  Do we already have this article in our index?  Change thread from
 *  -2 (ART_EXPIRED) to -1 (ART_NORMAL) if so and skip the header eating.
 */

		if ((respnum = valid_artnum (art)) >= 0 || art <= last_read_article) {
			arts[respnum].thread = ART_NORMAL;
			arts[respnum].unread = ART_UNREAD;
			continue;
		}

		if (! modified) {
			modified = TRUE;   /* we've modified the index */
							   /* it will need to be re-written */
		}

		if ((fd = open_header_fd (group_path, art)) < 0) {
			continue;
		}
		
		/*
		 *  Add article to arts[]
		 */
		if (top >= max_art)
			expand_art();

		arts[top].artnum = art;
		arts[top].thread = ART_NORMAL;

		set_article (&arts[top]);

		if (! parse_headers (fd, &arts[top])) {
			continue;
		}
		close (fd);
		last_read_article = arts[top].artnum;	/* used if arts are killed */
		top++;

		if (++count % 10 == 0 && ! update) {
#ifndef SLOW_SCREEN_UPDATE
			if (display_groupname) {
				MoveCursor (0, 0);
				CleartoEOLN ();
				center_line (0, TRUE, group);
				display_groupname = FALSE;
			}
			sprintf (msg, txt_indexing_num, count);
			wait_message (msg);
#else
			wait_message (txt_indexing);
#endif
		}
	}

	return modified;
}


/*
 *  Go through the articles in arts[] and use .thread to snake threads
 *  through them.  Use the subject line to construct threads.  The
 *  first article in a thread should have .inthread set to FALSE, the
 *  rest TRUE.  Only do unexprired articles we haven't visited yet
 *  (arts[].thread == -1 ART_NORMAL).
 */

void make_threads (rethread)
	int rethread;
{
	extern int cur_groupnum;
	register int i;
	register int j;

	/*
	 *  .thread & .inthread need to be reset if re-threading arts[]
	 */
	if (rethread && thread_arts &&
		active[my_group[cur_groupnum]].thread) {
		for (i=0 ; i < top ; i++) {
			arts[i].thread = ART_NORMAL;
			arts[i].inthread = FALSE;
		}
	}

	switch (sort_art_type) {
		case SORT_BY_NOTHING:		/* don't sort at all */
			qsort (arts, top, sizeof (struct header), artnum_comp);
			break;
		case SORT_BY_SUBJ_DESCEND:
		case SORT_BY_SUBJ_ASCEND:
			qsort (arts, top, sizeof (struct header), subj_comp);
			break;
		case SORT_BY_FROM_DESCEND:
		case SORT_BY_FROM_ASCEND:
			qsort (arts, top, sizeof (struct header), from_comp);
			break;
		case SORT_BY_DATE_DESCEND:
		case SORT_BY_DATE_ASCEND:
			qsort (arts, top, sizeof (struct header), date_comp);
			break;
		default:
			break;
	}

	if (thread_arts && active[my_group[cur_groupnum]].thread) {
		for (i = 0; i < top; i++) {
			if (arts[i].thread == ART_NORMAL) {
			    for (j = i+1; j < top; j++) {
					if (arts[j].thread != ART_EXPIRED &&
						((arts[i].subject == arts[j].subject) ||
						((arts[i].part || arts[i].patch) &&
						arts[i].archive == arts[j].archive))) {
							arts[i].thread = j;
							arts[j].inthread = TRUE;
							break;
					}
				}
			}
		}
	}
}

/*
 *  Return a pointer into s eliminating any leading Re:'s.  Example:
 *
 *	  Re: Reorganization of misc.jobs
 *	  ^   ^
 */

char *eat_re (s)
	char *s;
{

	while (*s == 'r' || *s == 'R') {
		if ((*(s+1) == 'e' || *(s+1) == 'E')) {
			if (*(s+2) == ':')
				s += 3;
			else if (*(s+2) == '^' && isdigit(*(s+3)) && *(s+4) == ':')
				s += 5;			/* hurray nn */
			else
				break;
		} else
			break;
		while (*s == ' ')
			s++;
	}

	return s;
}

/*
 *  Hash the subjects (after eating the Re's off) for a quicker
 *  thread search later.  We store the hashes for subjects in the
 *  index file for speed.
 */

long hash_s (s)
	char *s;
{
	long h = 0;
	unsigned char *t = (unsigned char *) s;

	while (*t)
		h = h * 64 + *t++;

	return h;
}


int parse_headers (fd, h)
	int fd;
	struct header *h;
{
	char buf[HEADER_LEN];
	char buf2[HEADER_LEN];
	char *ptr, *ptrline, *s;
	int n = 0, len = 0, lineno = 0;
	int flag;
	int got_subject = FALSE;
	int got_from = FALSE;
	int got_date = FALSE;
	int got_archive = FALSE;
	
	if ((n = read (fd, buf, HEADER_LEN)) <= 0) {
		return FALSE;
	}

	buf[n-1] = '\0';

	ptr = buf;

	while (1) {
		for (ptrline = ptr; *ptr && *ptr != '\n'; ptr++) {
			if (((*ptr) & 0xFF) < ' ') {
				*ptr = ' ';
			}
		}
		flag = *ptr;
		*ptr++ = '\0';
		lineno++;

		if (! got_from && match_header(ptrline, "From", buf2, HEADER_LEN)) {
			h->from = hash_str (buf2);
			got_from = TRUE;
		} else if (! got_subject && match_header(ptrline, "Subject", buf2, HEADER_LEN)) {
			s = eat_re (buf2);
			h->subject = hash_str (eat_re (s));
			got_subject = TRUE;
		} else if (! got_date && match_header(ptrline, "Date", buf2, HEADER_LEN)) {
			parse_date (buf2, h->date);
			got_date = TRUE;
		} else if (match_header(ptrline, "Archive-name", buf2, HEADER_LEN)) {
			if ((s = (char *) strchr (buf2, '/')) != NULL) {
				if (strncmp (s+1,"part",4) == 0 ||
				    strncmp (s+1,"Part",4) == 0) {
					h->part = str_dup (s+5);
					len = (int) strlen (h->part);
					if (h->part[len-1] == '\n') {
						h->part[len-1] = '\0';
					}
				} else {
					if (strncmp (s+1,"patch",5) == 0 ||
					    strncmp (s+1,"Patch",5) == 0) {
						h->patch = str_dup (s+6);
						len = (int) strlen (h->patch);
						if (h->patch[len-1] == '\n') {
							h->patch[len-1] = '\0';
						}
					}
				}
				if (h->part || h->patch) {
					s = buf2;
					while (*s && *s != '/')
						s++;
					*s = '\0';	
					s = buf2;
					h->archive = hash_str (s);
					got_archive = TRUE;
				}
			}
		}

		if (! flag || lineno > 25 || got_archive) {
			debug_print_header (h);
			return TRUE;
		}
	}
	/* NOTREACHED */
}

/* 
 *  Write out an index file.  Write the group name first so if
 *  local indexing is done we can disambiguate between group name
 *  hash collisions by looking at the index file.
 *
 *  NOTE: check out the add_string routine in hashstr.c to
 *  understand what *iptr is doing in this routine.
 */

void dump_index (group, killed)
	char *group;
	int killed;
{
	char nam[LEN];
	FILE *fp;
	int *iptr;
	int realnum;
	register int i;

	sprintf (nam, "%s.%d", index_file, getpid());
	if ((fp = fopen (nam, "w")) == NULL) {
		error_message (txt_cannot_open, nam);
		return;
	}

	/*
	 *  dump group header info.
	 */
	if (sort_art_type != SORT_BY_NOTHING) {
		qsort (arts, top, sizeof (struct header), artnum_comp);
	}
	fprintf(fp, "%s\n", group);
	fprintf(fp, "%d\n", num_of_arts ());
	if (last_read_article > arts[top-1].artnum) {
		fprintf(fp, "%ld\n", last_read_article);
	} else {
		fprintf(fp, "%ld\n", arts[top-1].artnum);
	}
	if (index_file_killed && killed) {
		fprintf (fp, "KILLED\n");
	} else {
		fprintf (fp, "COMPLETE\n");
	}

	/*
	 *  dump articles
	 */
	realnum = 0; 
	for (i = 0; i < top; i++) {
	    if (arts[i].thread != ART_EXPIRED && ! arts[i].tagged) { 
			debug_print_header (&arts[i]);

			fprintf(fp, "%ld\n", arts[i].artnum);

			iptr = (int *) arts[i].subject;
			iptr--;

			if (! arts[i].subject) {
				fprintf(fp, " \n");
			} else if (*iptr < 0 || *iptr > top) {
				fprintf(fp, " %s\n", arts[i].subject);
				*iptr = realnum;
/*
			} else if (arts[*iptr].tagged) {
				fprintf(fp, " %s\n", arts[i].subject);
				*iptr = realnum;
			} else if (killed && *iptr == i) {
*/
			} else if (killed || *iptr == i) {
				fprintf(fp, " %s\n", arts[i].subject);
			} else {
				fprintf(fp, "%%%d\n", *iptr);
			}
	
			iptr = (int *) arts[i].from;
			iptr--;

			if (! arts[i].from) {
				fprintf (fp, " \n");
			} else if (*iptr < 0 || *iptr > top) {
				fprintf (fp, " %s\n", arts[i].from);
				*iptr = realnum;
/*
			} else if (arts[*iptr].tagged) {
				fprintf(fp, " %s\n", arts[i].from);
				*iptr = realnum;
			} else if (killed && *iptr == i) {
*/			
			} else if (killed || *iptr == i) {
				fprintf(fp, " %s\n", arts[i].from);
			} else {
				fprintf(fp, "%%%d\n", *iptr);
			}

			fprintf (fp, "%s\n", arts[i].date);
			
			iptr = (int *) arts[i].archive;
			iptr--;

			if (! arts[i].archive) {
				fprintf (fp, "\n");
			} else if (*iptr < 0 || *iptr > top) {
				fprintf (fp, " %s\n", arts[i].archive);
				*iptr = realnum;
/*
			} else if (arts[*iptr].tagged) {
				fprintf (fp, " %s\n", arts[i].archive);
				*iptr = realnum;
*/				
			} else if (arts[i].part || arts[i].patch) {
/*
				if (killed && *iptr == i) {
*/				
				if (killed || *iptr == i) {
					fprintf(fp, " %s\n", arts[i].archive);
				} else {
					fprintf (fp, "%%%d\n", *iptr);
				}
			} else {
				fprintf (fp, "\n");
			}
			
			if (! arts[i].part) {
				fprintf (fp, " \n");
			} else {
				fprintf (fp, "%s\n", arts[i].part);
			}

			if (! arts[i].patch) {
				fprintf (fp, " \n");
			} else {
				fprintf (fp, "%s\n", arts[i].patch);
			}

			realnum++;
		}
	}
	fclose (fp);
	chmod (index_file, 0644);
	rename_file (nam, index_file);
	if (debug) {
		sprintf (msg, "/bin/cp %s INDEX", index_file);
		system (msg);
	}
}

/*
 *  strncpy that stops at a newline and null terminates
 */

void my_strncpy(p, q, n)
	char *p;
	char *q;
	int n;
{
	while (n--) {
		if (! *q || *q == '\n')
			break;
		*p++ = *q++;
	}
	*p = '\0';
}

/*
 *  Read in an index file.
 *
 *  index file header 
 *    1.  newsgroup name (ie. alt.sources)
 *    2.  number of articles (ie. 26)
 *    3.  number of last read article (ie. 210)
 *    4.  Is this a complete/killed index file (ie. COMPLETE/KILLED)
 *  index file record
 *    1.  article number   (ie. 183)                [mandatory]
 *    2.  Subject: line    (ie. Which newsreader?)  [mandatory]
 *    3.  From: line       (ie. iain@norisc)        [mandatory]
 *    4.  Date: of posting (ie. 911231125959)       [mandatory]
 *    5.  Archive: name    (ie. compiler)           [optional]
 *    6.  Part number of Archive: name  (ie. 01)    [optional]
 *    7.  Patch number of Archive: name (ie. 01)    [optional]
 */

int load_index ()
{
	int error = 0;
	int i, n;
	char buf[LEN], *p;
	FILE *fp;

	top = 0;
	last_read_article = 0L;

	if ((fp = fopen (index_file, "r")) == NULL) {
		return FALSE;
	}

	debug_print_comment ("*** LOADING ***");

	/*
	 *  load header - discard group name, num. of arts in index file after any arts were killed
	 */
	if (fgets(buf, sizeof buf, fp) == NULL ||
		fgets(buf, sizeof buf, fp) == NULL) {
		error = 0;			
		goto corrupt_index;	
	}
	i = atoi (buf);

	/*
	 * num. of last_read_article including any that were killed
	 */
	if (fgets(buf, sizeof buf, fp) == NULL) {
		error = 1;				
		goto corrupt_index;	
	}							
	last_read_article = (long) atol (buf);
	
	/*
	 * is index file complete or were articles killed when it was dumped
	 */
	if (fgets(buf, sizeof buf, fp) == NULL) {
		error = 2;				
		goto corrupt_index;	
	}
	index_file_killed = (buf[0] == 'K' ? TRUE : FALSE);
	
	/*
	 *  load articles
	 */
	for (; top < i ; top++) {
		if (top >= max_art) {
			expand_art ();
		}

		arts[top].thread = ART_EXPIRED;
		set_article (&arts[top]);

		/*
		 * Article no.
		 */
		if (fgets(buf, sizeof buf, fp) == NULL) {
			error = 3;
			goto corrupt_index;
		}
		arts[top].artnum = (long) atol (buf);

		/*
		 * Subject:
		 */
		if (fgets(buf, sizeof buf, fp) == NULL) {
			error = 4;
			goto corrupt_index;
		}

		if (buf[0] == '%') {
			n = atoi (&buf[1]);
			if (n >= top || n < 0) {
				error = 5;
				goto corrupt_index;
			}
			arts[top].subject = arts[n].subject;
		} else if (buf[0] == ' ') {
			for (p = &buf[1];  *p && *p != '\n'; p++)
				continue;	
			*p = '\0';
			arts[top].subject = hash_str (&buf[1]);
		} else {
			error = 6;
			goto corrupt_index;
		}
			
		/*
		 * From:
		 */
		if (fgets(buf, sizeof buf, fp) == NULL) {
			error = 7;
			goto corrupt_index;
		}

		if (buf[0] == '%') {
			n = atoi (&buf[1]);
			if (n >= top || n < 0) {
				error = 8;
				goto corrupt_index;
			}
			arts[top].from = arts[n].from;
		} else if (buf[0] == ' ') {
			for (p = &buf[1];  *p && *p != '\n'; p++)
				continue;
			*p = '\0';
			arts[top].from = hash_str (&buf[1]);
		} else {
			error = 9;
			goto corrupt_index;
		}

		/*
		 * Date:
		 */
		if (fgets(buf, sizeof buf, fp) == NULL) {
			error = 10;
			goto corrupt_index;
		}

		buf[strlen (buf)-1] = '\0';
		my_strncpy (arts[top].date, buf, 12);

		/*
		 * Archive-name:
		 */
		if (fgets(buf, sizeof buf, fp) == NULL) {
			error = 11;
			goto corrupt_index;
		}

		if (buf[0] == '\n') {
			arts[top].archive = (char *) 0;
		} else if (buf[0] == '%') {
			n = atoi (&buf[1]);
			if (n > top || n < 0) {
				error = 12;
				goto corrupt_index;
			}
			arts[top].archive = arts[n].archive;
		} else if (buf[0] == ' ') {
			for (p = &buf[1]; *p && *p != '\n' ; p++)
				continue;
			*p = '\0';
			arts[top].archive = hash_str (&buf[1]);
		} else {
			error = 13;
			goto corrupt_index;
		}

		/*
		 * part no.
		 */
		if (fgets(buf, sizeof buf, fp) == NULL) {
			error = 14;
			goto corrupt_index;
		}

		if (buf[0] != ' ') { 
			buf[strlen (buf)-1] = '\0';
			arts[top].part = str_dup (buf);
		}

		/*
		 * patch no.
		 */
		if (fgets(buf, sizeof buf, fp) == NULL) {
			error = 15;
			goto corrupt_index;
		}

		if (buf[0] != ' ') { 
			buf[strlen (buf)-1] = '\0';
			arts[top].patch = str_dup (buf);
		}

		debug_print_header (&arts[top]);
	}

	fclose(fp);
	return TRUE;

corrupt_index:
	if (! update) {
		sprintf (msg, txt_corrupt_index, index_file, error, top); 
		error_message (msg, "");
	}

	if (debug) {
		sprintf (msg, "cp %s INDEX.BAD", index_file);
		system (msg);
	}

	unlink (index_file);
	top = 0;
	return FALSE;
}


/*
 *  Look in the local $HOME/RCDIR/INDEXDIR (or wherever) directory for the
 *  index file for the given group.  Hashing the group name gets
 *  a number.  See if that #.1 file exists; if so, read first line.
 *  Group we want?  If no, try #.2.  Repeat until no such file or
 *  we find an existing file that matches our group.
 */

void find_local_index (group)
	char *group;
{
	unsigned long h;
	static char buf[LEN];
	int i;
	char *p;
	FILE *fp;

	h = hash_groupname (group);

	i = 1;
	while (1) {
		sprintf(index_file, "%s/%lu.%d", indexdir, h, i);

		if ((fp = fopen(index_file, "r")) == NULL) {
			return;
		}

		if (fgets(buf, sizeof buf, fp) == NULL) {
			fclose(fp);
			return;
		}
		fclose(fp);

		for (p = buf; *p && *p != '\n'; p++)
			continue;
		*p = '\0';

		if (strcmp(buf, group) == 0)
			return;

		i++;
	}
}

/*
 *  Run the index file updater only for the groups we've loaded.
 */

void do_update()
{
	int i, j;
	char group_path[LEN];
	char *p;
	long epoch;
	
	if (verbose) {
		time (&epoch);
		printf ("%s", ctime (&epoch));
		fflush (stdout);
	}

	for (i = 0; i < local_top; i++) {
		strcpy(group_path, active[my_group[i]].name);
		for (p = group_path; *p; p++) {
			if (*p == '.') {
				*p = '/';
			}
		}
		if (verbose) {
			printf ("%s %s\n", (catchup ? "Catchup" : "Updating"),
					active[my_group[i]].name);
			fflush (stdout);
		}
		index_group (active[my_group[i]].name, group_path);
		if (catchup) {
			for (j = 0; j < top; j++) {
				arts[j].unread = ART_READ;
			}
			update_newsrc (active[my_group[i]].name, my_group[i], FALSE);
		}
	}

	if (verbose) {
		time (&epoch);
		printf ("%s", ctime (&epoch));
		fflush (stdout);
	}
}

/*
 *  Check for articles and say how many new/unread in each group.
 *  or
 *  Start if new/unread articles and return first group with new/unread.
 *  or
 *  Save any new articles to savedir and mark arts read and mail user
 *  and inform how many arts in which groups were saved.
 */

int check_start_save_any_news (check_start_save)
	int check_start_save;
{
	char buf[LEN], log[LEN], *p;
	char group_path[LEN];
	extern FILE *note_fp;
	FILE *fp, *fp_log;
	int i, j, print_group;
	int check_arts = 0;
	int	log_opened = TRUE;
	int print_first = TRUE;
	int saved_arts = 0;
	int saved_groups = 0;
	int unread_news = FALSE;	
	long epoch;

	switch (check_start_save) {
		case CHECK_ANY_NEWS:
			if (verbose) {
				wait_message (txt_checking_for_news);
			}
			break;
		case START_ANY_NEWS:
			wait_message (txt_checking_for_news);
			break;
		case SAVE_ANY_NEWS:
			sprintf (log, "%s/log", rcdir);
			if ((fp_log = fopen (log, "w")) == NULL) {
				error_message (txt_cannot_open, log);
				fp_log = stdout;
				verbose = FALSE;
				log_opened = FALSE;
			}
			time (&epoch);
			fprintf (fp_log, "To: %s\r\n", userid);
			fprintf (fp_log, "Subject: NEWS LOG %s\r\n", ctime (&epoch));
			break;
	}
	
	for (i = 0; i < local_top; i++) {
		strcpy (group_path, active[my_group[i]].name);
		for (p = group_path; *p; p++) {
			if (*p == '.') {
				*p = '/';
			}
		}
		
		index_group (active[my_group[i]].name, group_path);
		read_newsrc_line (active[my_group[i]].name);
		print_group = TRUE;
		check_arts = 0;

		for (j = 0; j < top; j++) {
			if (arts[j].unread == ART_UNREAD)  {
				switch (check_start_save) {
					case CHECK_ANY_NEWS:
						if (print_first && verbose) {
							putchar ('\n');
							print_first = FALSE;
						}
						check_arts++;
						break;
					case START_ANY_NEWS:
						return i;	/* return first group with unread news */ 
						/* NOTREACHED */
					case SAVE_ANY_NEWS:
						if (print_group) {	
							sprintf (msg, "Saved %s...\n", active[my_group[i]].name);
							fprintf (fp_log, "%s", msg);
							if (verbose) {
								wait_message (msg);
							}
							print_group = FALSE;
							saved_groups++;
							sprintf (buf, "%s/dummy", group_path);
							create_path (buf);
						}
						sprintf (msg, "[%5ld]  %s\n", arts[j].artnum, arts[j].subject);
						fprintf (fp_log, "%s", msg);
						if (verbose) {
							wait_message (msg);
						}
						saved_arts++;

						sprintf (buf, "%s/%s/%ld", savedir, group_path, arts[j].artnum);
						if ((fp = fopen (buf, "w")) == NULL) {
							fprintf (fp_log, txt_cannot_open, buf);
							error_message (txt_cannot_open, buf);
							continue;
						}
				
						art_open (arts[j].artnum, group_path);	
						fseek (note_fp, 0L, 0);
						copy_fp (note_fp, fp, (char *) 0);
						art_close ();
						fclose (fp);
						break;
				}
			}
		}
		
		if (check_start_save == SAVE_ANY_NEWS) {
			if (catchup) {
				for (j = 0; j < top; j++) {
					arts[j].unread = ART_READ;
				}
				update_newsrc (active[my_group[i]].name, my_group[i], FALSE);
			}
		} else {
			if (check_arts) {
				if (verbose) {
					sprintf (msg, "%4d unread articles in %s\n",
						check_arts, active[my_group[i]].name);
					wait_message (msg); 	
				}
				unread_news = TRUE;	
			}
		}
	}
	switch (check_start_save) {
		case CHECK_ANY_NEWS:
			if (unread_news) {
				return 2;
			} else {
				if (verbose) {
					wait_message (txt_there_is_no_news);
				}
				return 0;
			}
			/* NOTREACHED */ 
		case START_ANY_NEWS:
			wait_message (txt_there_is_no_news);
			return -1;
			/* NOTREACHED */ 
		case SAVE_ANY_NEWS:
			sprintf (msg, "\nSaved %d article(s) from %d group(s)\n", saved_arts, saved_groups);
			fprintf (fp_log, "%s", msg);
			if (verbose) {
				wait_message (msg);
			}
			if (log_opened) {
				fclose (fp_log);
				if (verbose) {
					sprintf (msg, "Mailing log to %s\n", userid);
					wait_message (msg);
				}
				sprintf (buf, "%s \"%s\" < %s", mailer, userid, log);
				if (! invoke_cmd (buf)) {
					error_message (txt_command_failed_s, buf);
				}
			}
			/* NOTREACHED */ 
	}
	return 0;
}

/*
 *  reload index after any articles have been killed
 */
 
void reload_index_file (group, killed)
	char *group;
	int killed;
{
	char group_path[LEN];
	char *p;

	if (local_index) {			/* writing index in home directory */
		set_real_uid_gid ();	/* so become them */
	}

	strcpy (group_path, group);			/* turn comp.unix.amiga into */
	for (p = group_path; *p; p++)		/* comp/unix/amiga */
		if (*p == '.')
			*p = '/';

	if (killed) {
		if (! update) {
			wait_message (txt_killing_arts);
		}
		index_file_killed = TRUE;
		setup_base (group, group_path);
		dump_index (group, killed);
		load_index ();
	} else {
		if (! update) {
			wait_message (txt_unkilling_arts);
		}
		if (local_index) {
			find_local_index (group);
		} else {
			sprintf (index_file, "%s/%s/%s", spooldir, group_path, INDEXDIR);
		}

		unlink (index_file);	/* delete index file */

		index_file_killed = FALSE;
		last_read_article = 0L;

		if (read_group (group, group_path)) {
			dump_index (group, killed);
		}
	}

	make_threads (TRUE);
	find_base ();

	if (local_index) {
		set_tin_uid_gid ();
	}

	return; 
}

/*
 * convert date from "24 Jul 91 12:59:59" to "910724125959"
 */

char *parse_date (date, str)
	char *date;
	char *str;
{
	char buf[4];
	int i = 3;

	if (date[1] == ' ') {	/* ie. "2 Aug..." instead of "12 Aug... */
		str[4] = '0';		/* day */
		str[5] = date[0];
		i = 2;
	} else {
		str[4] = date[0];		/* day */
		str[5] = date[1];
	}
	
	buf[0] = date[i++];		/* month in Jan,Feb,.. form */
	buf[1] = date[i++];
	buf[2] = date[i++];
	buf[3] = '\0';

	i++;
	
	str[0] = date[i++];		/* year */
	str[1] = date[i++];
	
	i++;
	
	if (strcmp (buf, "Jan") == 0) {		/* convert Jan to 01 etc */
		str[2] = '0';
		str[3] = '1';
	} else if (strcmp (buf, "Feb") == 0) {
		str[2] = '0';
		str[3] = '2';
	} else if (strcmp (buf, "Mar") == 0) {
		str[2] = '0';
		str[3] = '3';
	} else if (strcmp (buf, "Apr") == 0) {
		str[2] = '0';
		str[3] = '4';
	} else if (strcmp (buf, "May") == 0) {
		str[2] = '0';
		str[3] = '5';
	} else if (strcmp (buf, "Jun") == 0) {
		str[2] = '0';
		str[3] = '6';
	} else if (strcmp (buf, "Jul") == 0) {
		str[2] = '0';
		str[3] = '7';
	} else if (strcmp (buf, "Aug") == 0) {
		str[2] = '0';
		str[3] = '8';
	} else if (strcmp (buf, "Sep") == 0) {
		str[2] = '0';
		str[3] = '9';
	} else if (strcmp (buf, "Oct") == 0) {
		str[2] = '1';
		str[3] = '0';
	} else if (strcmp (buf, "Nov") == 0) {
		str[2] = '1';
		str[3] = '1';
	} else if (strcmp (buf, "Dec") == 0) {
		str[2] = '1';
		str[3] = '2';
	} else {
		str[2] = '0';
		str[3] = '0';
	}
	
	str[6] = date[i++];		/* hour */
	str[7] = date[i++];

	i++;
	
	str[8] = date[i++];		/* minutes */
	str[9] = date[i++];
	
	i++;
	
	str[10] = date[i++];	/* seconds */
	str[11] = date[i++];

	str[12] = '\0';		/* terminate string */

	return (str);
}


int artnum_comp (p1, p2)
	char *p1;
	char *p2;
{
	struct header *s1 = (struct header *) p1;
	struct header *s2 = (struct header *) p2;

	/* s1->artnum less than s2->artnum */
	if (s1->artnum < s2->artnum) {
		return -1;
	}
	/* s1->artnum greater than s2->artnum */
	if (s1->artnum > s2->artnum) {
		return 1;
	}
	return 0;
}


int subj_comp (p1, p2)
	char *p1;
	char *p2;
{
	struct header *s1 = (struct header *) p1;
	struct header *s2 = (struct header *) p2;

	/* s1->subject less than s2->subject */
	if (strcmp (s1->subject, s2->subject) < 0) {
		return (sort_art_type == SORT_BY_SUBJ_DESCEND ? -1 : 1);
	}
	/* s1->subject greater than s2->subject */
	if (strcmp (s1->subject, s2->subject) > 0) {
		return (sort_art_type == SORT_BY_SUBJ_DESCEND ? 1 : -1);
	}
	return 0;
}


int from_comp (p1, p2)
	char *p1;
	char *p2;
{
	struct header *s1 = (struct header *) p1;
	struct header *s2 = (struct header *) p2;

	/* s1->from less than s2->from */
	if (strcmp (s1->from, s2->from) < 0) {
		return (sort_art_type == SORT_BY_FROM_DESCEND ? -1 : 1);
	}
	/* s1->from greater than s2->from */
	if (strcmp (s1->from, s2->from) > 0) {
		return (sort_art_type == SORT_BY_FROM_DESCEND ? 1 : -1);
	}
	return 0;
}


int date_comp (p1, p2)
	char *p1;
	char *p2;
{
	struct header *s1 = (struct header *) p1;
	struct header *s2 = (struct header *) p2;

	/* s1->date less than s2->date */
	if (strcmp (s1->date, s2->date) < 0) {
		return (sort_art_type == SORT_BY_DATE_DESCEND ? -1 : 1);
	}
	/* s1->date greater than s2->date */
	if (strcmp (s1->date, s2->date) > 0) {
		return (sort_art_type == SORT_BY_DATE_DESCEND ? 1 : -1);
	}
	return 0;
}


void set_article (art)
	struct header *art;
{	
	art->subject = (char *) 0;
	art->from = (char *) 0;
	art->date[0] = '\0';
	art->archive = (char *) 0;
	art->part = (char *) 0;
	art->patch = (char *) 0;
	art->unread = ART_UNREAD;
	art->inthread = FALSE;
	art->tagged = FALSE;
}
